@@ -1,8 +1,20 @@
 """Interface for the joplin data API."""
 
+import copy
 import json
 import logging
-from typing import Any, Dict, List, Optional
+from typing import (
+    Any,
+    Callable,
+    cast,
+    Dict,
+    List,
+    MutableMapping,
+    Optional,
+    TypedDict,
+    Union,
+)
+import urllib.parse
 
 import requests
 
@@ -12,10 +24,51 @@ logging.getLogger("requests").setLevel(logging.WARNING)
 logging.getLogger("urllib3").setLevel(logging.WARNING)
 
 
+##############################################################################
+# Typing support.
+##############################################################################
+
+
+# Datatypes used by the Joplin API. Needed for arbitrary kwargs.
+JoplinTypes = Union[float, int, str]
+# Kwargs mapping of the datatypes.
+JoplinKwargs = MutableMapping[str, JoplinTypes]
+
+
+class JoplinItem(TypedDict, total=False):
+    id: str
+    body: str
+    title: str
+    parent_id: str
+    created_time: int
+    updated_time: int
+    type_: int
+    # event specific
+    type: int
+    item_type: int
+    item_id: str
+    # resource specific
+    mime: str
+    filename: str
+    file_extension: str
+    size: int
+
+
+class JoplinItemList(TypedDict, total=False):
+    items: List[JoplinItem]
+    has_more: bool
+    cursor: int
+
+
+##############################################################################
+# Base wrapper that manages the requests to the REST API.
+##############################################################################
+
+
 class ApiBase:
     """Contains the basic requests of the REST API."""
 
-    def __init__(self, token: str, url: str = "http://localhost:41184"):
+    def __init__(self, token: str, url: str = "http://localhost:41184") -> None:
         self.url = url
         self.token = token
 
@@ -23,22 +76,24 @@ class ApiBase:
         self,
         method: str,
         path: str,
-        query: Optional[dict] = None,
-        data: Optional[dict] = None,
-        **kwargs,
+        query: Optional[JoplinKwargs] = None,
+        data: Optional[JoplinKwargs] = None,
+        files: Optional[Dict[str, Any]] = None,
     ) -> requests.models.Response:
-        logging.debug(f"API: {method} request: {path=}, {query=}, {data=}, {kwargs=}")
-        if data is not None:
+        logging.debug(f"API: {method} request: {path=}, {query=}, {data=}, {files=}")
+        if data is not None and "id_" in data:
             # "id" is a reserved keyword in python, so don't use it.
-            data["id"] = data.pop("id_", None)
+            data["id"] = data["id_"]
         if query is None:
             query = {}
         query["token"] = self.token  # TODO: extending the dict may have side effects
         query_str = "&".join([f"{key}={val}" for key, val in query.items()])
 
         try:
-            response = getattr(requests, method)(
-                f"{self.url}{path}?{query_str}", json=data, **kwargs
+            response: requests.models.Response = getattr(requests, method)(
+                f"{self.url}{path}?{query_str}",
+                json=data,
+                files=files,
             )
             logging.debug(f"API: response {response.text}")
             response.raise_for_status()
@@ -47,21 +102,30 @@ class ApiBase:
             raise
         return response
 
-    def delete(self, *args) -> requests.models.Response:
+    def delete(self, path: str) -> requests.models.Response:
         """Convenience method to issue a delete request."""
-        return self._request("delete", *args)
+        return self._request("delete", path)
 
-    def get(self, *args, **kwargs) -> requests.models.Response:
+    def get(
+        self, path: str, query: Optional[JoplinKwargs] = None
+    ) -> requests.models.Response:
         """Convenience method to issue a get request."""
-        return self._request("get", *args, **kwargs)
+        return self._request("get", path, query=query)
 
-    def post(self, *args, **kwargs) -> requests.models.Response:
+    def post(
+        self,
+        path: str,
+        data: Optional[JoplinKwargs] = None,
+        files: Optional[Dict[str, Any]] = None,
+    ) -> requests.models.Response:
         """Convenience method to issue a post request."""
-        return self._request("post", *args, **kwargs)
+        return self._request("post", path, data=data, files=files)
 
-    def put(self, *args, **kwargs) -> requests.models.Response:
+    def put(
+        self, path: str, data: Optional[JoplinKwargs] = None
+    ) -> requests.models.Response:
         """Convenience method to issue a put request."""
-        return self._request("put", *args, **kwargs)
+        return self._request("put", path, data=data)
 
 
 ##############################################################################
@@ -71,39 +135,47 @@ class ApiBase:
 
 
 class Event(ApiBase):
-    def get_event(self, id_: str, **kwargs):
-        """Get the event with the given ID."""
-        return self.get(f"/events/{id_}", query=kwargs).json()
+    """
+    Events are supported since Joplin 2.4.5.
+    See: https://github.com/laurent22/joplin/releases/tag/v2.4.5
+    """
 
-    def get_events(self, **kwargs):
+    def get_event(self, id_: str, **query: JoplinTypes) -> JoplinItem:
+        """Get the event with the given ID."""
+        response: JoplinItem = self.get(f"/events/{id_}", query=query).json()
+        return response
+
+    def get_events(self, **query: JoplinTypes) -> JoplinItemList:
         """
         Get events, paginated. To get all events (unpaginated), use
         "get_all_events()".
         """
-        return self.get("/events", query=kwargs).json()
+        response: JoplinItemList = self.get("/events", query=query).json()
+        return response
 
 
 class Note(ApiBase):
-    def add_note(self, **kwargs) -> str:
+    def add_note(self, **data: JoplinTypes) -> str:
         """Add a note."""
-        response = self.post("/notes", data=kwargs)
-        return response.json()["id"]
+        response: JoplinItem = self.post("/notes", data=data).json()
+        return response["id"]
 
-    def delete_note(self, id_: str):
+    def delete_note(self, id_: str) -> None:
         """Delete a note."""
         self.delete(f"/notes/{id_}")
 
-    def get_note(self, id_: str, **kwargs):
+    def get_note(self, id_: str, **query: JoplinTypes) -> JoplinItem:
         """Get the note with the given ID."""
-        return self.get(f"/notes/{id_}", query=kwargs).json()
+        response: JoplinItem = self.get(f"/notes/{id_}", query=query).json()
+        return response
 
     def get_notes(
         self,
         notebook_id: Optional[str] = None,
         resource_id: Optional[str] = None,
         tag_id: Optional[str] = None,
-        **kwargs,
-    ):
+        **query: JoplinTypes,
+    ) -> JoplinItemList:
         """
         Get notes, paginated. If a notebook, resource or tag ID is given,
         return the corresponding notes. To get all notes (unpaginated), use
@@ -114,37 +186,42 @@ class Note(ApiBase):
         notebook = "" if notebook_id is None else f"/folders/{notebook_id}"
         resource = "" if resource_id is None else f"/resources/{resource_id}"
         tag = "" if tag_id is None else f"/tags/{tag_id}"
-        return self.get(f"{notebook}{resource}{tag}/notes", query=kwargs).json()
+        response: JoplinItemList = self.get(
+            f"{notebook}{resource}{tag}/notes", query=query
+        ).json()
+        return response
 
-    def modify_note(self, id_: str, **kwargs):
+    def modify_note(self, id_: str, **data: JoplinTypes) -> None:
         """Modify a note."""
-        self.put(f"/notes/{id_}", data=kwargs)
+        self.put(f"/notes/{id_}", data=data)
 
 
 class Notebook(ApiBase):
-    def add_notebook(self, **kwargs) -> str:
+    def add_notebook(self, **data: JoplinTypes) -> str:
         """Add a notebook."""
-        response = self.post("/folders", data=kwargs)
-        return response.json()["id"]
+        response: JoplinItem = self.post("/folders", data=data).json()
+        return response["id"]
 
-    def delete_notebook(self, id_: str):
+    def delete_notebook(self, id_: str) -> None:
         """Delete a notebook."""
         self.delete(f"/folders/{id_}")
 
-    def get_notebook(self, id_: str, **kwargs):
+    def get_notebook(self, id_: str, **query: JoplinTypes) -> JoplinItem:
         """Get the notebook with the given ID."""
-        return self.get(f"/folders/{id_}", query=kwargs).json()
+        response: JoplinItem = self.get(f"/folders/{id_}", query=query).json()
+        return response
 
-    def get_notebooks(self, **kwargs):
+    def get_notebooks(self, **query: JoplinTypes) -> JoplinItemList:
         """
         Get notebooks, paginated. To get all notebooks (unpaginated), use
         "get_all_notebooks()".
         """
-        return self.get("/folders", query=kwargs).json()
+        response: JoplinItemList = self.get("/folders", query=query).json()
+        return response
 
-    def modify_notebook(self, id_: str, **kwargs):
+    def modify_notebook(self, id_: str, **data: JoplinTypes) -> None:
         """Modify a notebook."""
-        self.put(f"/folders/{id_}", data=kwargs)
+        self.put(f"/folders/{id_}", data=data)
 
 
 class Ping(ApiBase):
@@ -154,81 +231,94 @@ class Ping(ApiBase):
 
 
 class Resource(ApiBase):
-    def add_resource(self, filename: str, **kwargs) -> str:
+    def add_resource(self, filename: str, **data: JoplinTypes) -> str:
         """Add a resource."""
         # Preserve the filename if there is no title specified.
-        if kwargs.get("title") is None:
-            kwargs["title"] = filename
+        if data.get("title") is None:
+            data["title"] = filename
         with open(filename, "rb") as infile:
             files = {
                 "data": (json.dumps(filename), infile),
-                "props": (None, json.dumps(kwargs)),
+                "props": (None, json.dumps(data)),
             }
-            response = self.post("/resources", files=files)
-        return response.json()["id"]
+            response: JoplinItem = self.post("/resources", files=files).json()
+        return response["id"]
 
-    def delete_resource(self, id_: str):
+    def delete_resource(self, id_: str) -> None:
         """Delete a resource."""
         self.delete(f"/resources/{id_}")
 
-    def get_resource(self, id_: str, get_file: bool = False, **kwargs):
+    def get_resource(
+        self, id_: str, get_file: bool = False, **query: JoplinTypes
+    ) -> JoplinItem:
         """Get the resource with the given ID."""
         file_ = "/file" if get_file else ""
-        return self.get(f"/resources/{id_}{file_}", query=kwargs).json()
+        response: JoplinItem = self.get(f"/resources/{id_}{file_}", query=query).json()
+        return response
 
-    def get_resources(self, note_id: Optional[str] = None, **kwargs):
+    def get_resources(
+        self, note_id: Optional[str] = None, **query: JoplinTypes
+    ) -> JoplinItemList:
         """
         Get resources, paginated. If a note ID is given, return the corresponding
         resources. To get all resources (unpaginated), use "get_all_resources()".
         """
         note = "" if note_id is None else f"/notes/{note_id}"
-        return self.get(f"{note}/resources", query=kwargs).json()
+        response: JoplinItemList = self.get(f"{note}/resources", query=query).json()
+        return response
 
-    def modify_resource(self, id_: str, **kwargs):
+    def modify_resource(self, id_: str, **data: JoplinTypes) -> None:
         """Modify a resource."""
-        self.put(f"/resources/{id_}", data=kwargs)
+        self.put(f"/resources/{id_}", data=data)
 
 
 class Search(ApiBase):
-    def search(self, **kwargs):
+    def search(self, **query: JoplinTypes) -> JoplinItemList:
         """Issue a search."""
-        response = self.get("/search", query=kwargs)
-        return response.json()
+        # Copy the dict, because its content gets changed.
+        outer_query = copy.deepcopy(query)
+        # The outer query is the query of the URL. The inner query is the query of the
+        # Joplin search endpoint. To avoid special characters in the inner query that
+        # would be recognized as outer query, encode it.
+        outer_query["query"] = urllib.parse.quote(cast(str, outer_query["query"]))
+        response: JoplinItemList = self.get("/search", query=outer_query).json()
+        return response
 
 
 class Tag(ApiBase):
-    def add_tag(self, tag_id: Optional[str] = None, **kwargs) -> str:
+    def add_tag(self, tag_id: Optional[str] = None, **data: JoplinTypes) -> str:
         """
         Add a tag. If a tag is given, add the tag to a note.
         The data has to contain the note ID.
         """
         note = "" if tag_id is None else f"/{tag_id}/notes"
-        response = self.post(f"/tags{note}", data=kwargs)
-        return response.json()["id"]
+        response: JoplinItem = self.post(f"/tags{note}", data=data).json()
+        return response["id"]
 
-    def delete_tag(self, id_: str, note_id: Optional[str] = None):
+    def delete_tag(self, id_: str, note_id: Optional[str] = None) -> None:
         """Delete a tag. If a note is given, remove the tag from this note."""
         note = "" if note_id is None else f"/notes/{note_id}"
         self.delete(f"/tags/{id_}{note}")
 
-    def get_tag(self, id_: str, **kwargs):
+    def get_tag(self, id_: str, **query: JoplinTypes) -> JoplinItem:
         """Get the tag with the given ID."""
-        return self.get(f"/tags/{id_}", query=kwargs).json()
+        response: JoplinItem = self.get(f"/tags/{id_}", query=query).json()
+        return response
 
-    def get_tags(self, note_id: Optional[str] = None, **kwargs):
+    def get_tags(
+        self, note_id: Optional[str] = None, **query: JoplinTypes
+    ) -> JoplinItemList:
         """
         Get tags, paginated. If a note is given, return the corresponding tags.
         To get all tags (unpaginated), use "get_all_tags()".
         """
         note = "" if note_id is None else f"/notes/{note_id}"
-        return self.get(f"{note}/tags", query=kwargs).json()
+        response: JoplinItemList = self.get(f"{note}/tags", query=query).json()
+        return response
 
-    def modify_tag(self, id_: str, **kwargs):
+    def modify_tag(self, id_: str, **data: JoplinTypes) -> None:
         """Modify a tag."""
-        self.put(f"/tags/{id_}", data=kwargs)
-
-
-ITEM = Dict[str, Any]
+        self.put(f"/tags/{id_}", data=data)
 
 
 class Api(Event, Note, Notebook, Ping, Resource, Search, Tag):
@@ -237,12 +327,12 @@ class Api(Event, Note, Notebook, Ping, Resource, Search, Tag):
     This should be the only class accessed from the users.
     """
 
-    def add_tag_to_note(self, tag_id: str, note_id: str):
+    def add_tag_to_note(self, tag_id: str, note_id: str) -> None:
         """Add a tag to a given note."""
         note = self.get_note(id_=note_id, fields="id")
         self.add_tag(tag_id=tag_id, id_=note["id"])
 
-    def add_resource_to_note(self, resource_id: str, note_id: str):
+    def add_resource_to_note(self, resource_id: str, note_id: str) -> None:
         """Add a resource to a given note."""
         note = self.get_note(id_=note_id, fields="body")
         resource = self.get_resource(id_=resource_id, fields="title")
@@ -251,7 +341,7 @@ class Api(Event, Note, Notebook, Ping, Resource, Search, Tag):
         )
         self.modify_note(note_id, body=body_with_attachment)
 
-    def delete_all_notebooks(self):
+    def delete_all_notebooks(self) -> None:
         """Delete all notebooks."""
         notebooks = self.get_notebooks()["items"]
         for notebook in notebooks:
@@ -259,46 +349,53 @@ class Api(Event, Note, Notebook, Ping, Resource, Search, Tag):
             if not notebook["parent_id"]:
                 self.delete_notebook(notebook["id"])
 
-    def delete_all_resources(self):
+    def delete_all_resources(self) -> None:
         """Delete all resources."""
         resources = self.get_resources()["items"]
         for resource in resources:
             self.delete_resource(resource["id"])
 
-    def delete_all_tags(self):
+    def delete_all_tags(self) -> None:
         """Delete all tags."""
         tags = self.get_tags()["items"]
         for tag in tags:
             self.delete_tag(tag["id"])
 
     @staticmethod
-    def _get_all(func, **kwargs):
+    def _unpaginate(
+        func: Callable[..., JoplinItemList], **query: JoplinTypes
+    ) -> List[JoplinItem]:
         """Calls an Joplin endpoint until it's response doesn't contain more data."""
-        response = func(**kwargs)
+        response: JoplinItemList = func(**query)
         items = response["items"]
         page = 1  # pages are one based
         while response["has_more"]:
             page += 1
-            response = func(page=page, **kwargs)
+            query["page"] = page
+            response = func(**query)
             items.extend(response["items"])
         return items
 
-    def get_all_events(self, **kwargs) -> List[ITEM]:
+    def get_all_events(self, **query: JoplinTypes) -> List[JoplinItem]:
         """Get all events, unpaginated."""
-        return self._get_all(self.get_events, **kwargs)
+        return self._unpaginate(self.get_events, **query)
 
-    def get_all_notes(self, **kwargs) -> List[ITEM]:
+    def get_all_notes(self, **query: JoplinTypes) -> List[JoplinItem]:
         """Get all notes, unpaginated."""
-        return self._get_all(self.get_notes, **kwargs)
+        return self._unpaginate(self.get_notes, **query)
 
-    def get_all_notebooks(self, **kwargs) -> List[ITEM]:
+    def get_all_notebooks(self, **query: JoplinTypes) -> List[JoplinItem]:
         """Get all notebooks, unpaginated."""
-        return self._get_all(self.get_notebooks, **kwargs)
+        return self._unpaginate(self.get_notebooks, **query)
 
-    def get_all_resources(self, **kwargs) -> List[ITEM]:
+    def get_all_resources(self, **query: JoplinTypes) -> List[JoplinItem]:
         """Get all resources, unpaginated."""
-        return self._get_all(self.get_resources, **kwargs)
+        return self._unpaginate(self.get_resources, **query)
 
-    def get_all_tags(self, **kwargs) -> List[ITEM]:
+    def get_all_tags(self, **query: JoplinTypes) -> List[JoplinItem]:
         """Get all tags, unpaginated."""
-        return self._get_all(self.get_tags, **kwargs)
+        return self._unpaginate(self.get_tags, **query)
+
+    def search_all(self, **query: JoplinTypes) -> List[JoplinItem]:
+        """Issue a search and get all results, unpaginated."""
+        return self._unpaginate(self.search, **query)
