@@ -10,18 +10,20 @@ from re import sub
 import string
 import tempfile
 import time
+from typing import cast, Iterable, Tuple
 import unittest
 
 from parameterized import parameterized
 import requests
 
-from joppy.api import Api
+from joppy.api import Api, JoplinKwargs
 from joppy import tools
 from . import setup_joplin
 
 
+os.makedirs("test_output", exist_ok=True)
 logging.basicConfig(
-    filename="test.log",
+    filename="test_output/test.log",
     filemode="w",
     format="%(asctime)s [%(levelname)s]: %(message)s",
     level=logging.DEBUG,
@@ -101,7 +103,6 @@ class TestBase(unittest.TestCase):
         self.api.delete_all_notebooks()
         self.api.delete_all_resources()
         self.api.delete_all_tags()
-        self.current_cursor = self.api.get_events()["cursor"]
 
     @staticmethod
     def get_random_id() -> str:
@@ -139,7 +140,9 @@ class TestBase(unittest.TestCase):
         return 0 <= timestamp <= int(time.time() * 1000)  # ms
 
     @staticmethod
-    def get_combinations(iterable, max_combinations: int = 100):
+    def get_combinations(
+        iterable: Iterable[str], max_combinations: int = 100
+    ) -> Iterable[Tuple[str, ...]]:
         """Get some combinations of an iterable."""
         # https://stackoverflow.com/a/10465588
         # TODO: Randomize fully. For now the combinations are sorted by length.
@@ -150,7 +153,7 @@ class TestBase(unittest.TestCase):
             itertools.combinations(list_, r)
             for r in lengths
             # shuffle each iteration
-            if random.shuffle(list_) is None  # type: ignore
+            if random.shuffle(list_) is None  # type: ignore [func-returns-value]
         )
         return itertools.islice(combinations, max_combinations)
 
@@ -164,16 +167,17 @@ class Event(TestBase):
     ]
     default_properties = ["id", "item_type", "item_id", "type", "created_time"]
 
-    def generate_event(self):
+    def generate_event(self) -> None:
         """Generate an event and wait until it's available."""
-        events_before = len(self.api.get_all_events(cursor=self.current_cursor))
+        current_cursor = self.api.get_events()["cursor"]
+        events_before = len(self.api.get_all_events(cursor=current_cursor))
         # For now only notes trigger events:
         # https://joplinapp.org/api/references/rest_api/#events
         self.api.add_notebook()
         self.api.add_note()
 
         def compare_event_count():
-            event_count = len(self.api.get_all_events(cursor=self.current_cursor))
+            event_count = len(self.api.get_all_events(cursor=current_cursor))
             return event_count if event_count == events_before + 1 else None
 
         # Wait until the event is available.
@@ -181,19 +185,21 @@ class Event(TestBase):
 
     def test_get_event(self):
         """Get a specific event."""
+        current_cursor = self.api.get_events()["cursor"]
         self.generate_event()
-        last_event = self.api.get_events(cursor=self.current_cursor)["items"][-1]
+        last_event = self.api.get_events(cursor=current_cursor)["items"][-1]
         event = self.api.get_event(id_=last_event["id"])
 
         self.assertEqual(list(event.keys()), self.default_properties + ["type_"])
-        self.assertIn(event["item_type"], ItemType._value2member_map_)
+        self.assertIn(event["item_type"], {item.value for item in ItemType})
         self.assertTrue(self.is_id_valid(event["item_id"]))
         self.assertEqual(event["type"], ChangeType.CREATED.value)
         self.assertTrue(self.is_timestamp_valid(event["created_time"]))
         self.assertEqual(event["type_"], ItemType.ITEM_CHANGE.value)
 
         for property_ in self.default_properties:
-            self.assertEqual(event[property_], event[property_])
+            # https://github.com/python/mypy/issues/7178
+            self.assertEqual(event[property_], event[property_])  # type: ignore [misc]
 
     def test_get_events_by_cursor(self):
         """Get all events by specifying a cursor of 0."""
@@ -203,10 +209,13 @@ class Event(TestBase):
         previous_created_time = 0
         previous_id = 0
         for event in events["items"]:
-            self.assertGreater(event["created_time"], previous_created_time)
-            self.assertGreater(event["id"], previous_id)
-            previous_created_time = event["created_time"]
-            previous_id = event["id"]
+            # TODO: id is int only for events. Find a way to abstract it.
+            current_id = cast(int, event["id"])
+            current_created_time = event["created_time"]
+            self.assertGreater(current_created_time, previous_created_time)
+            self.assertGreater(current_id, previous_id)
+            previous_created_time = current_created_time
+            previous_id = current_id
             self.assertEqual(list(event.keys()), self.default_properties)
 
     def test_get_events_empty(self):
@@ -460,7 +469,7 @@ class Resource(TestBase):
     default_properties = ["id", "title"]
 
     @with_resource
-    def test_add(self, filename: str):
+    def test_add(self, filename):
         """Add a resource."""
         id_ = self.api.add_resource(filename=filename)
         self.assertTrue(self.is_id_valid(id_))
@@ -470,7 +479,7 @@ class Resource(TestBase):
         self.assertEqual(resources[0]["id"], id_)
 
     @with_resource
-    def test_add_to_note(self, filename: str):
+    def test_add_to_note(self, filename):
         """Add a resource to an existing note."""
         self.api.add_notebook()
         note_id = self.api.add_note()
@@ -488,7 +497,7 @@ class Resource(TestBase):
         # self.assertEqual(notes[0]["id"], note_id)
 
     @with_resource
-    def test_delete(self, filename: str):
+    def test_delete(self, filename):
         """Add and then delete a resource."""
         id_ = self.api.add_resource(filename=filename)
         resources = self.api.get_resources()
@@ -539,7 +548,7 @@ class Resource(TestBase):
                 self.assertEqual(list(resource.keys()), list(properties))
 
     @with_resource
-    def test_modify_title(self, filename: str):
+    def test_modify_title(self, filename):
         """Modify a resource title."""
         id_ = self.api.add_resource(filename=filename)
 
@@ -562,7 +571,7 @@ class Resource(TestBase):
             self.assertEqual(resource["size"], os.path.getsize(file_))
 
     @with_resource
-    def test_check_property_title(self, filename: str):
+    def test_check_property_title(self, filename):
         """Check the title of a resource."""
         title = self.get_random_string()
         id_ = self.api.add_resource(filename=filename, title=title)
@@ -614,7 +623,7 @@ class Search(TestBase):
     def test_master_key(self):
         """There is no master key configured."""
         self.assertEqual(
-            self.api.search(query={"query": "*", "type": "master_key"}),
+            self.api.search(query="*", type="master_key"),
             self.empty_search,
         )
 
@@ -636,7 +645,7 @@ class Search(TestBase):
         for _ in range(limit + 1):
             self.api.add_notebook()
 
-        query = {"query": "*", "type": "folder", "limit": limit}
+        query: JoplinKwargs = {"query": "*", "type": "folder", "limit": limit}
         search_result = self.api.search(**query)
         self.assertEqual(len(search_result["items"]), limit)
         self.assertTrue(search_result["has_more"])
@@ -646,6 +655,42 @@ class Search(TestBase):
         self.assertEqual(len(search_result["items"]), 1)
         self.assertFalse(search_result["has_more"])
 
+    def test_search_query_special_chars(self):
+        """
+        Search should succeed even with special characters.
+        See: https://github.com/marph91/joppy/issues/5
+        """
+        queries = (
+            (
+                "https://books.google.com.br/books?id=vaZFBgAAQBAJ&pg=PA83&dq=lakatos+"
+                "copernicus&hl=pt-BR&sa=X&ved=0ahUKEwjewoWZ6q7hAhUNJ7kGHdy5CZUQ6AEIUDA"
+                r"F#v=onepage&q=lakatos%20copernicus&f=false"
+            ),
+            (
+                "https://www.facebook.com/photo.php?fbid=621934757910715&set=a.1437947"
+                "52391387&type=3&app=fbl"
+            ),
+            r"foo# ?bar%+!baz",
+            self.get_random_string(),
+        )
+
+        for query in queries:
+            self.api.add_notebook(title=query)
+            result = self.api.search(query=query, type="folder")
+            self.assertEqual(len(result["items"]), 1)
+            self.assertEqual(result["items"][0]["title"], query)
+
+    def test_search_all(self):
+        """Search notebooks and return all results, unpaginated."""
+        # Small limit and count to create/remove as less as possible items.
+        count, limit = random.randint(1, 10), random.randint(1, 10)
+        title = self.get_random_string()
+        query: JoplinKwargs = {"query": title, "type": "folder", "limit": limit}
+        for _ in range(count):
+            self.api.add_notebook(title=title)
+        self.assertEqual(len(self.api.search(**query)["items"]), min(limit, count))
+        self.assertEqual(len(self.api.search_all(**query)), count)
+
 
 class Tag(TestBase):
     properties = [
@@ -803,7 +848,7 @@ class Helper(TestBase):
 
 class Miscellaneous(TestBase):
     @with_resource
-    def test_same_id_different_type(self, filename: str):
+    def test_same_id_different_type(self, filename):
         """Same IDs can be used if the types are different."""
         id_ = self.get_random_id()
 
